// SeaProto class implementation
// Copyright (C) 04.09.97 Marco Dettori and Oliver Obst
//
// $Id: seaproto.C,v 0.2 1998/03/16 10:41:18 oliver Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

#include <iostream.h>
#include <strings.h>
#include "unistd.h"

#include <qsocknot.h>

#include "seaproto.h"

// --- constructor ------------------------------------------------------------
// SeaProto :: SeaProto
//              const char      *hostname       : name of the server
//              int             port            : which port number to use

SeaProto :: SeaProto( const char *hostname, int portNr,
                      QObject *parent, const char *name )
        : QObject( parent, name )
{
        server = new char[ strlen( hostname ) + 1 ];
        strcpy( server, hostname );
        port = portNr;
        server_info = 0;
        sn = 0;
#ifdef HAVE_TLI
        callp = 0;
#endif
}

// --- destructor -------------------------------------------------------------
// SeaProto :: ~SeaProto

SeaProto :: ~SeaProto()
{
        delete server;
        delete server_info;
        delete sn;
#ifdef HAVE_TLI
        delete callp;
#endif        
}

// --- public -----------------------------------------------------------------
// SeaProto :: setServer

void SeaProto :: setServer( const char *hostname )
{
        delete server;
        server  = new char[ strlen( hostname ) + 1 ];
        strcpy( server, hostname );
}

// --- public -----------------------------------------------------------------
// SeaProto :: setPort

void SeaProto :: setPort( int portNr )
{
        port = portNr;
}


// --- public -----------------------------------------------------------------
// SeaProto :: seaOpen
// 
// opens connection to server

int SeaProto :: seaOpen( char *opname, char *myname )
{
        char    buffer[ 100 ];
        
        connection = TRUE;

        delete server_info;
        if (( server_info = gethostbyname( server )) == NULL ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: gethostbyname: "
                     << server << " not found\n";
                return -1;
        }

        memset((char *) &epAddr, 0, sizeof( epAddr ));
        epAddr.sin_family       = AF_INET;
        memcpy((char *) &epAddr.sin_addr.s_addr, server_info->h_addr,
               server_info->h_length );
        epAddr.sin_port         = htons( u_short( port ));

#ifdef HAVE_TLI
        if (( sockfd = t_open( DEV_TCP, O_RDWR, 0 )) < 0 ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: t_open error\n";
                return -1;
        }

        if ( t_bind( sockfd, 0, 0 ) < 0 ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: t_bind error\n";
                return -1;
        }

        delete callp;
        if (( callp = ( t_call * ) t_alloc( sockfd, T_CALL, T_ADDR )) == 0 ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: t_alloc error\n";
                return -1;
        }

        callp->addr.maxlen = sizeof( epAddr );
        callp->addr.len    = sizeof( epAddr );
        callp->addr.buf    = (char *) &epAddr;
        callp->opt.len     = 0;
        callp->udata.len   = 0;

        if ( t_connect( sockfd, callp, 0 ) < 0 ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: t_connect error\n";
                return -1;
        }

        t_snd( sockfd, myname, strlen( myname ) + 1, 0 );
        int flags;
        int n = t_rcv( sockfd, buffer, 100, &flags );
        strncpy( opname, buffer, n - 2 );

#else   // HAVEN'T TLI, USE SOCKETS
        if (( sockfd = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: create socket error\n";
                return -1;
        }

        if ( ::connect( sockfd, (struct sockaddr *) &epAddr , sizeof( epAddr ))
             < 0 ) {
                connection = FALSE;
                cerr << "seaproto.C: seaOpen: connect socket error\n";
                return -1;
        }

        write( sockfd, myname, strlen( myname ) + 1 );
        int n = read( sockfd, buffer, 100 );
        strncpy( opname, buffer, n - 2 );

#endif  HAVE_TLI        

        delete sn;
        sn = new QSocketNotifier( sockfd, QSocketNotifier::Read, this );
        sn->setEnabled( TRUE );
        QObject::connect( sn, SIGNAL( activated(int) ), SLOT( readSocket() ) );

        if ( n > 1 && buffer[ n - 2 ] == char( 0xFF ) ) {
                if ( buffer[ n - 1 ] == ETX ) 
                        return 0;
                if ( buffer[ n - 1 ] == ETB )
                        return 1;
        }

        return -1;
}

// --- public -----------------------------------------------------------------
//
// seaProto :: seaClose( bool surrender )
//
// close the connection

void SeaProto :: seaClose( bool surrender )
{
        char    buffer[ 2 ];

        buffer[ 0 ] = 0xFF;
        buffer[ 1 ] = EOT;
        
        if ( surrender )
#ifdef HAVE_TLI
                t_snd( sockfd, buffer, 2, 0 );
#else        
                write( sockfd, buffer, 2 );
#endif                

        disconnect( sn );
        delete sn;
        connection = FALSE;

#ifdef HAVE_TLI
        t_close( sockfd );
#else
        close( sockfd );
#endif
}


// --- public -----------------------------------------------------------------
//
// seaProto :: connected()
//
// return true if connection to server established

bool SeaProto :: connected()
{
        return connection;
}


// --- public slot ------------------------------------------------------------
// seaProto:: writeReport( int id, int result )
//
// send ID and result via TCP to opponent

void SeaProto :: writeReport( int id, int result )
{
        char     buffer[ 5 ];

        buffer[ 0 ] = id / 256;
        buffer[ 1 ] = id % 256;
        buffer[ 2 ] = result % 256;
        buffer[ 3 ] = 0xFF;

        switch( result ) {
        case MISSED : 
                buffer[ 4 ] = ETX;
                break;
        case DEFEAT :
                buffer[ 4 ] = EOT;
                break;
        default :
                buffer[ 4 ] = ETB;
        }

#ifdef HAVE_TLI
        t_snd( sockfd, buffer, 5, 0 );
#else        
        write( sockfd, buffer, 5 );
#endif
}

// --- public slot ------------------------------------------------------------
// seaProto::writePeek( int id );
//
// send request for a field

void SeaProto :: writePeek( int id )
{
        char    buffer[ 4 ];

        buffer[ 0 ] = id / 256;
        buffer[ 1 ] = id % 256;
        buffer[ 2 ] = 0xFF;
        buffer[ 3 ] = ETB;

#ifdef HAVE_TLI
        t_snd( sockfd, buffer, 4, 0 );
#else
        write( sockfd, buffer, 4 );
#endif        
}

// --- protected slot ---------------------------------------------------------
// SeaProto::readSocket()
//
// read socket and send the result via signal

void SeaProto :: readSocket()
{
        disconnect( sn );
        delete sn;
        sn = new QSocketNotifier( sockfd, QSocketNotifier::Read, this );
        sn->setEnabled( TRUE );
        QObject::connect( sn, SIGNAL( activated(int) ), SLOT( readSocket() ) );

        char    buffer[ 100 ];

#ifdef HAVE_TLI
        int flags;
        int r = t_rcv( sockfd, buffer, sizeof( buffer ), &flags );
#else        
        int r = read( sockfd, buffer, sizeof( buffer ));
#endif        

        switch ( r ) {
        case 4 :
                emit request( buffer[ 0 ] * 256 + buffer[ 1 ] );
                break;
        case 5 :
                emit reporter( buffer[ 0 ] * 256 + buffer[ 1 ], buffer[ 2 ] );
                break;
        default :
                emit defeat();
        }
}
